from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

from typing_extensions import Annotated, Doc

from dymmond_settings import __version__
from dymmond_settings.enums import EnvironmentType


@dataclass
class BaseSettings:
    """
    Base of all the settings for any system.
    """

    def dict(self, exclude_none: bool = False, upper: bool = False) -> Dict[str, Any]:
        """
        Dumps all the settings into a python dictionary.
        """
        original = asdict(self)

        if not exclude_none:
            if not upper:
                return original
            return {k.upper(): v for k, v in original.items()}

        if not upper:
            return {k: v for k, v in original.items() if v is not None}
        return {k.upper(): v for k, v in original.items() if v is not None}

    def tuple(self, exclude_none: bool = False, upper: bool = False) -> List[Tuple[str, Any]]:
        """
        Dumps all the settings into a tuple.
        """
        original = asdict(self)

        if not exclude_none:
            if not upper:
                return list(original.items())
            return list({k.upper(): v for k, v in original.items()}.items())

        if not upper:
            return [(k, v) for k, v in original.items() if v is not None]
        return [(k.upper(), v) for k, v in original.items() if v is not None]


@dataclass
class Settings(BaseSettings):
    debug: Annotated[
        bool,
        Doc(
            """
            Boolean indicating if the application should return the debug tracebacks on
            server errors, in other words, if you want to have debug errors being displayed.

            !!! Tip
                Do not use this in production as `True`.
            """
        ),
    ] = field(default=False)
    environment: Annotated[
        Optional[str],
        Doc(
            """
            Optional string indicating the environment where the settings are running.
            You won't probably need this but it is here in case you might want to use.
            """
        ),
    ] = field(default=EnvironmentType.PRODUCTION)
    version: Annotated[
        Union[str, int, float],
        Doc(
            """
            The version of the application and defaults to the current version of the settings
            system if not set.
            """
        ),
    ] = field(default=__version__)



import os
from typing import TYPE_CHECKING, Any, Optional, Type

from dymmond_settings.functional import LazyObject, empty
from dymmond_settings.module_loading import import_string

if TYPE_CHECKING:
    from dymmond_settings.base import Settings

OVERRIDE_VARIABLE = "OVERRIDE_SETTINGS_MODULE_VARIABLE"
ENVIRONMENT_VARIABLE = os.environ.get(OVERRIDE_VARIABLE) or "SETTINGS_MODULE"


class LazySettings(LazyObject):
    """
    A lazy proxy for either global application settings or a custom settings object.
    The user can manually configure settings prior to using them. Otherwise,
    The system uses the settings module pointed to by SETTINGS_MODULE.
    """

    def _setup(self, name: Optional[str] = None) -> None:
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time settings are needed, if the user hasn't
        configured settings manually.
        """
        settings_module: str = os.environ.get(
            ENVIRONMENT_VARIABLE, "dymmond_settings.global_settings.Settings"
        )

        settings: Type["Settings"] = import_string(settings_module)

        for setting, _ in settings().dict().items():
            assert setting.islower(), "%s should be in lower case." % setting

        self._wrapped = settings()

    def __repr__(self: "LazySettings") -> str:
        # Hardcode the class name as otherwise it yields 'Settings'.
        if self._wrapped is empty:
            return "<LazySettings [Unevaluated]>"
        return '<LazySettings "{settings_module}">'.format(
            settings_module=self._wrapped.__class__.__name__
        )

    @property
    def configured(self) -> Any:
        """Return True if the settings have already been configured."""
        return self._wrapped is not empty


settings: Type["Settings"] = LazySettings()  # type: ignore
